from typing import Literal as _Literal


Action = _Literal[
    "auth",
    "pay",
    "hold",
    "payqr",
    "paytoken",
    "paycash",
    "hold_completion",
    "subscribe",
    "paydonate",
    "paysplit",
    "split_rules",
    "letter_of_credit",
    "apay",
    "gpay",
    "regular",
    "refund",
    "payment_prepare",
    "p2pcredit",
    "p2pdebit",
    "confirm",
    "mpi",
    "cardverification",
    "register",
    "data",
    "ticket",
    "status",
    "invoice_send",
    "invoice_cancel",
    "token_create",
    "token_create_unique",
    "token_update",
    "reports",
    "reports_compensation",
    "reports_compensation_file",
    "agent_shop_create",
    "agent_shop_edit",
    "agent_info_merchant",
    "agent_info_user",
    "unsubscribe",
    "subscribe_update",
]



from functools import singledispatchmethod
from dataclasses import asdict

from base64 import b64encode
from json import JSONEncoder

from uuid import UUID
from decimal import Decimal
from datetime import date, datetime, UTC

from liqpy.models.request import FiscalItem, DetailAddenda, SplitRule
from liqpy.constants import DATE_FORMAT


__all__ = ("Encoder", "JSONEncoder", "SEPARATORS")


SEPARATORS = (",", ":")


class Encoder(JSONEncoder):
    """Custom JSON encoder for LiqPay API requests"""

    date_fmt = DATE_FORMAT

    tz = UTC

    def __init__(self) -> None:
        super().__init__(
            skipkeys=False,
            ensure_ascii=True,
            check_circular=True,
            allow_nan=False,
            sort_keys=False,
            indent=None,
            separators=SEPARATORS,
            default=None,
        )

    @singledispatchmethod
    def default(self, o):
        return super().default(o)

    @default.register
    def _(self, o: Decimal) -> float:
        return round(float(o), 4)

    @default.register
    def _(self, o: datetime) -> str:
        return o.astimezone(self.tz).strftime(self.date_fmt)

    @default.register
    def _(self, o: date) -> str:
        return o.strftime(self.date_fmt)

    @default.register
    def _(self, o: bytes) -> str:
        return o.decode("utf-8")

    @default.register
    def _(self, o: UUID) -> str:
        return str(o)

    @default.register
    def _(self, o: DetailAddenda) -> str:
        return b64encode(self.encode(o.to_json()).encode()).decode()

    @default.register
    def _(self, o: SplitRule) -> dict:
        return asdict(o)

    @default.register
    def _(self, o: FiscalItem) -> dict:
        return asdict(o)



from pandas import (
    read_csv as pd_read_csv,
    UInt64Dtype,
    Float64Dtype,
    CategoricalDtype,
    StringDtype,
    DataFrame,
    NA,
)

from liqpy.models.report import Currency, Status, Action, Code, PayWay
from liqpy.constants import DATE_FORMAT

ID_TYPE = UInt64Dtype()
STRING_TYPE = StringDtype()
FLOAT_TYPE = Float64Dtype()

CURRENCY_TYPE = CategoricalDtype(categories=Currency._member_map_.values())
ACTION_TYPE = CategoricalDtype(categories=Action._member_map_.values())

STATUS_TYPE = CategoricalDtype(categories=Status._member_map_.values())
CODE_TYPE = CategoricalDtype(categories=Code._member_map_.values())
PAYWAY_TYPE = CategoricalDtype(categories=PayWay._member_map_.values())

COLUMN_TYPES = {
    "ID": ID_TYPE,
    "AMOUNT": FLOAT_TYPE,
    "SENDER_COMMISSION": FLOAT_TYPE,
    "RECEIVER_COMMISSION": FLOAT_TYPE,
    "AGENT_COMMISSION": FLOAT_TYPE,
    "CURRENCY": CURRENCY_TYPE,
    "AMOUNT_CREDIT": FLOAT_TYPE,
    "COMISSION_CREDIT": FLOAT_TYPE,
    "CURRENCY_CREDIT": CURRENCY_TYPE,
    "CREATE_DATE": STRING_TYPE,
    "END_DATE": STRING_TYPE,
    "TYPE": STRING_TYPE,
    "STATUS": STATUS_TYPE,
    "STATUS_ERR_CODE": CODE_TYPE,
    "AUTH_CODE": STRING_TYPE,
    "SHOP_ORDER_ID": STRING_TYPE,
    "DESCRIPTION": STRING_TYPE,
    "PHONE": STRING_TYPE,
    "SENDER_COUNTRY_CODE": STRING_TYPE,
    "CARD": STRING_TYPE,
    "ISSUER_BANK": STRING_TYPE,
    "CARD_COUNTRY": STRING_TYPE,
    "CARD_TYPE": STRING_TYPE,
    "PAY_WAY": PAYWAY_TYPE,
    "RECEIVER_CARD": STRING_TYPE,
    "RECEIVER_OKPO": ID_TYPE,
    "REFUND_AMOUNT": FLOAT_TYPE,
    "REFUND_DATE_LAST": STRING_TYPE,
    # "REFUND_RESERVE_IDS": STRING_TYPE,
    "RESERVE_REFUND_ID": ID_TYPE,
    "RESERVE_PAYMENT_ID": ID_TYPE,
    "RESERVE_AMOUNT": FLOAT_TYPE,
    "RESERVE_DATE": STRING_TYPE,
    "COMPLETION_DATE": STRING_TYPE,
    "INFO": STRING_TYPE,
    "LIQPAY_ORDER_ID": STRING_TYPE,
    "COMPENSATION_ID": ID_TYPE,
    "COMPENSATION_DATE": STRING_TYPE,
    "BONUSPLUS_ACCOUNT": STRING_TYPE,
    "BONUS_TYPE": STRING_TYPE,
    "BONUS_PERCENT": FLOAT_TYPE,
    "BONUS_AMOUNT": FLOAT_TYPE,
}

DATE_COLUMNS = [key for key in COLUMN_TYPES.keys() if "_DATE" in key]


def parse_refund_reserve_ids(x):
    if x:
        return frozenset(map(lambda v: int(v), x.split("|")))
    else:
        return NA


def read_liqpay_csv(source, **kwargs) -> DataFrame:
    """
    Read a CSV file from LiqPay with the appropriate column types

    Arguments
    ---------

    - `source`: File path or file-like object
    - `**kwargs`: Additional arguments to pass to `pandas.read_csv`
    """
    return pd_read_csv(
        source,
        encoding="utf-16",
        dtype=COLUMN_TYPES,
        parse_dates=DATE_COLUMNS,
        date_format=DATE_FORMAT,
        index_col="ID",
        converters={"REFUND_RESERVE_IDS": parse_refund_reserve_ids},
        **kwargs,
    )



from typing import TYPE_CHECKING, Any, AnyStr, Optional, Unpack

from enum import Enum
from uuid import UUID

from urllib.parse import urljoin
from base64 import b64encode, b64decode
from hashlib import sha1

from datetime import datetime

from liqpy.constants import URL, VERSION

from .encoder import Encoder, JSONEncoder, SEPARATORS
from .decoder import Decoder, JSONDecoder
from .preprocess import Preprocessor, BasePreprocessor
from .validation import Validator, BaseValidator
from .exceptions import exception

if TYPE_CHECKING:
    from requests import Session, Response

    from liqpy.types import LiqpayRequestDict
    from liqpy.types.action import Action
    from liqpy.types.post import Hooks, Proxies, Timeout, Verify, Cert


__all__ = [
    "Encoder",
    "Decoder",
    "Preprocessor",
    "Validator",
    "exception",
    "Endpoint",
    "is_sandbox",
    "post",
    "sign",
    "encode",
    "decode",
    "request",
]


class Endpoint(Enum):
    """LiqPay API endpoints"""

    REQUEST: str = "/api/request"
    CHECKOUT: str = f"/api/{VERSION}/checkout"

    def url(self) -> str:
        """Return full URL for the endpoint"""
        return urljoin(URL, self.value)


def is_sandbox(key: str, /) -> bool:
    """Check if the key is a sandbox key"""
    return key.startswith("sandbox_")


def post(
    endpoint: Endpoint,
    /,
    data: AnyStr,
    signature: AnyStr,
    *,
    session: "Session",
    stream: bool = False,
    allow_redirects: bool = False,
    proxies: Optional["Proxies"] = None,
    timeout: Optional["Timeout"] = None,
    hooks: Optional["Hooks"] = None,
    verify: Optional["Verify"] = None,
    cert: Optional["Cert"] = None,
) -> "Response":
    """
    Send POST request to LiqPay API

    See [Rules for the formation of a request for payment](https://www.liqpay.ua/en/documentation/data_signature).

    Arguments
    ---------
    - `endpoint` -- API endpoint to send request to (see `liqpy.Endpoint`)
    - `data` -- base64 encoded JSON data to send
    - `signature` -- LiqPay signature for the data
    - `session` -- `requests.Session` instance to use
    - `stream` -- whether to stream the response
    - `allow_redirects` -- whether to follow redirects
    - `proxies` -- proxies to use
    (see [Requests Proxies](https://docs.python-requests.org/en/stable/user/advanced/#proxies))
    - `timeout` -- timeout for the request
    - `hooks` -- hooks for the request
    (see [Requests Event Hooks](https://docs.python-requests.org/en/stable/user/advanced/#event-hooks))
    - `verify` -- whether to verify SSL certificate
    (see [Request SSL Cert Verification](https://requests.readthedocs.io/en/stable/user/advanced/#ssl-cert-verification))
    - `cert` -- client certificate to use
    (see [Request Client Side Certificates](https://requests.readthedocs.io/en/stable/user/advanced/#client-side-certificates))

    Returns
    -------
    - `requests.Response` instance

    Example
    -------
    >>> from requests import Session
    >>> from liqpy.api import encode, sign, request, Endpoint
    >>> data = encode({"action": "status", "version": 3})
    >>> signature = sign(data, key=b"a4825234f4bae72a0be04eafe9e8e2bada209255")
    >>> with Session() as session:  # doctest: +SKIP
    ...     response = request(Endpoint.REQUEST, data, signature, session=session) # doctest: +SKIP
    ...     result = response.json() # doctest: +SKIP
    """
    # print(data)
    response = session.request(
        method="POST",
        url=endpoint.url(),
        data={"data": data, "signature": signature},
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        json=None,
        params=None,
        cookies=None,
        files=None,
        auth=None,
        proxies=proxies,
        timeout=timeout,
        hooks=hooks,
        allow_redirects=allow_redirects,
        stream=stream,
        verify=verify,
        cert=cert,
    )
    response.raise_for_status()
    return response


def sign(data: bytes, /, key: bytes) -> bytes:
    """
    Sign data string with private key

    Algorithm:

    1. Concatenate the private key with the data string as `key + data + key`
    2. Calculate SHA1 hash of the concatenated string
    3. Encode the hash in base64

    >>> data = encode({"action": "status", "version": 3})
    >>> sign(data, key=b"a4825234f4bae72a0be04eafe9e8e2bada209255")
    b'qI0/snsDFB7MiYUxrqhBqX2420E='
    """
    return b64encode(sha1(key + data + key).digest())


def encode(
    params: "LiqpayRequestDict",
    /,
    *,
    filter_none: bool = True,
    encoder: Optional[JSONEncoder] = None,
    preprocessor: Optional[BasePreprocessor] = None,
    validator: Optional[BaseValidator] = None,
) -> bytes:
    """
    Encode parameters into base64 encoded JSON

    >>> encode({"action": "status", "version": 3})
    b'eyJhY3Rpb24iOiAic3RhdHVzIiwgInZlcnNpb24iOiAzfQ=='
    """
    if filter_none:
        params = {key: value for key, value in params.items() if value is not None}

    if encoder is None:
        encoder = JSONEncoder(separators=SEPARATORS)

    if preprocessor is not None:
        preprocessor(params, encoder=encoder)

    if validator is not None:
        validator(params)

    return b64encode(encoder.encode(params).encode())


def decode(data: bytes, /, decoder: Optional[JSONDecoder] = None) -> dict[str, Any]:
    """Decode base64 encoded JSON"""
    if decoder is None:
        decoder = JSONDecoder()

    return decoder.decode(b64decode(data).decode())


def request(
    action: "Action",
    /,
    public_key: str,
    *,
    version: int = VERSION,
    **params: "Unpack[LiqpayRequestDict]",
) -> "LiqpayRequestDict":
    """
    Create data dictionary for LiqPay API request

    >>> request("status", key="...", order_id="a1a1a1a1")
    {'action': 'status', 'public_key': '...', 'version': 3, 'order_id': 'a1a1a1a1'}
    """
    params.update(action=action, public_key=public_key, version=version)

    liqpay_id = params.pop("opid", None)
    if isinstance(liqpay_id, (str, UUID)):
        params.setdefault("order_id", liqpay_id)
    elif isinstance(liqpay_id, int):
        params.setdefault("payment_id", str(liqpay_id))

    match action:
        case "subscribe":
            assert "subscribe_periodicity" in params, "subscribe_periodicity is required"
            params["subscribe"] = True

            if params.get("subscribe_date_start") is None:
                params["subscribe_date_start"] = datetime.now()
        
        case "letter_of_credit":
            params["letter_of_credit"] = True

    return params
